<?php

/**
 * @file
 * Provides Drupal implementation of project release info parsing and batch API.
 */

/**
 * Regular expression to match PHP function names. (From Drupal 7)
 *
 * @see http://php.net/manual/en/language.functions.php
 */
define('DRUPAL_PHP_FUNCTION_PATTERN', '[a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*');

module_load_include('package.inc', 'project_dependency');

/**
 * Minimum API compatibility to process. We will not look at modules before
 * Drupal 6.
 */
define('PROJECT_DEPENDENCY_MINIMUM_DRUPAL_VERSION', 6);

/**
 * Directory for git repositories to be processed.
 *
 * These get checked out automatically, but the whole process goes faster
 * if it's already prepopulated with everything we need.
 *
 */
define('PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY', variable_get('project_dependency_sourcecode_directory', '/tmp/project_dependency_sourcecode_directory'));

/**
 * Path to the upstream git repository.
 *
 * If the source code repository of any project is not found locally, it will be
 * fetched from this path.
 */
define('PROJECT_DEPENDENCY_UPSTREAM_REPOSITORY_PATH', variable_get('project_dependency_upstream_repository_path', 'git://git.drupal.org/project'));

/**
 * Parse all the .info files.
 *
 * @param $info_files
 *   List of component .info files.
 *
 * @return
 *   Associative array of component information keyed by component name and
 *   merged with an array of defaults.
 */
function project_dependency_info_parse(array $info_files) {
  $defaults = array(
    'name' => 'Unknown',
    'description' => '',
    'dependencies' => array(),
    'test_dependencies' => array(),
  );

  $info = array();
  foreach ($info_files as $file) {
    $component = basename($file, '.info');
    $info[$component] = drupal_parse_info_file($file) + $defaults;

    // Change info keys to suite project_dependency_info.
    $info[$component]['title'] = $info[$component]['name'];
    $info[$component]['name'] = $component;
  }
  return $info;
}

/**
 * Get the project release nodes that contain a component and are compatible
 * with an API, ordered by most recent (highest nid)
 *
 * Sadly, the component may appear in more than one product, since there's no
 * guaranteed uniqueness of a submodule name. So we're just preferring the
 * most recent release node (highest nid).
 *
 * @param $component
 *   Component name.
 * @param $api_tid
 *   Core API compatibility tid.
 *
 * @return
 *   Array of release arrays, keyed by project release node nid, each of which
 *   has the keys
 *   - 'uri'           Project URI (shortname)
 *   - 'component_id'  Numeric key
 *   - 'pid'           Project nid
 *   - 'release_nid'
 *   - 'version'       Release name
 *   - 'tag'           VCS tag
 *   - 'version_major' project_release_node.version_major
 *   . Each indicates that the given component
 *   name is found in the given release. These are ordered by version string
 *   descending.
 */
function project_dependency_info_releases_for_component_get($component, $api_tid) {
  $result = db_query("SELECT DISTINCT pp.uri, pp.nid AS pid, c.component_id, c.release_nid, prn.version, prn.tag, prn.version_major
    FROM {project_dependency_component} c
    INNER JOIN {project_release_nodes} prn ON c.release_nid = prn.nid
    LEFT JOIN {project_projects} pp ON prn.pid = pp.nid
    INNER JOIN {term_node} t ON c.release_nid = t.nid
    WHERE t.tid = %d AND c.name = '%s'
    ORDER BY prn.nid DESC, t.vid DESC
  ", $api_tid, $component);

  $releases = array();
  while ($release = db_fetch_array($result)) {
    $releases[$release['release_nid']] = $release;
  }
  return $releases;
}

/**
 * Attempt to determine the Drupal core API term.
 *
 * @param $node
 *   Node object.
 *
 * @return
 *   Core API term to which the node belongs, otherwise FALSE.
 */
function project_dependency_info_core_api($node) {
  static $api_terms = array();

  if (empty($node->nid)) {
    return NULL;
  }
  if (empty($api_terms[$node->nid])) {
    $api_terms[$node->nid] = FALSE;

    if (!empty($node->taxonomy)) {
      // Relase API version vocabulary.
      $api_vid = _project_release_get_api_vid();

      foreach ($node->taxonomy as $tid => $term) {
        if ($term->vid == $api_vid) {
          $api_terms[$node->nid] = $term;
        }
      }
    }
  }
  return $api_terms[$node->nid];
}

/**
 * Given a core API taxonomy term name like '6.x', return the numeric core api
 * associated with it.
 *
 * @param $core_api_term_name
 */
function project_dependency_core_api_from_term($core_api_term_name) {
  return preg_replace('/\.x.*$/', '', $core_api_term_name);
}

/**
 * Parse a dependency for comparison by project_dependency_info_check_incompatibility().
 *
 * @param $dependency
 *   A dependency string, for example 'foo (>=7.x-4.5-beta5, 3.x)'.
 * @param $core_api
 *   The term name for the core version of Drupal. Like '6.x'
 *
 * @return
 *   An associative array with three keys:
 *   - 'name' includes the component name of the thing to depend on
 *     (e.g. 'page_manager').
 *   - 'original_version' contains the original version string (which can be
 *     used in the UI for reporting incompatibilities).
 *   - 'versions' is a list of associative arrays, each containing the keys
 *     'op' and 'version'. 'op' can be one of: '=', '==', '!=', '<>', '<',
 *     '<=', '>', or '>='. 'version' is one piece like '4.5-beta3'.
 *   Callers should pass this structure to project_dependency_info_check_incompatibility().
 *
 * @see project_dependency_info_check_incompatibility()
 */
function project_dependency_info_parse_dependency($dependency, $core_api) {
  // We use named subpatterns and support every op that version_compare
  // supports. Also, op is optional and defaults to equals.
  $p_op = '(?P<operation>!=|==|=|<|<=|>|>=|<>)?';
  // Core version is always optional: 7.x-2.x and 2.x is treated the same.
  $p_core = '(?:' . preg_quote($core_api . '.x') . '-)?';
  $p_major = '(?P<major>\d+)';
  // By setting the minor version to x, branches can be matched.
  $p_minor = '(?P<minor>(?:\d+|x)(?:-[A-Za-z]+\d+)?)';
  $value = array();
  $parts = explode('(', $dependency, 2);
  $value['name'] = trim($parts[0]);
  if (isset($parts[1])) {
    $value['original_version'] = ' (' . $parts[1];
    foreach (explode(',', $parts[1]) as $version) {
      if (preg_match("/^\s*$p_op\s*$p_core$p_major\.$p_minor/", $version, $matches)) {
        $op = !empty($matches['operation']) ? $matches['operation'] : '=';
        if ($matches['minor'] == 'x') {
          // Drupal considers "2.x" to mean any version that begins with
          // "2" (e.g. 2.0, 2.9 are all "2.x"). PHP's version_compare(),
          // on the other hand, treats "x" as a string; so to
          // version_compare(), "2.x" is considered less than 2.0. This
          // means that >=2.x and <2.x are handled by version_compare()
          // as we need, but > and <= are not.
          if ($op == '>' || $op == '<=') {
            $matches['major']++;
          }
          // Equivalence can be checked by adding two restrictions.
          if ($op == '=' || $op == '==') {
            $value['versions'][] = array('op' => '<', 'version' => ($matches['major'] + 1) . '.x');
            $op = '>=';
          }
        }
        $value['versions'][] = array('op' => $op, 'version' => $matches['major'] . '.' . $matches['minor']);
      }
    }
  }
  return $value;
}

/**
 * Check out a release and determine relevant information.
 *
 * Locate all .info files contained within the project, determine the names of
 * the module(s) contained within the project and store the dependencies for
 * later processing.
 *
 * @param $release
 *   A keyed array of release information, containing
 *   - 'tag' => The tag of the release (like 7.x-1.0)
 *   - 'uri' => The project shortname (like 'views')
 *   - 'nid' => The nid of the project release node
 *
 * @return
 *   An array of dependencies, or FALSE on failure.
 */
function project_dependency_info_batch_process_release(array $release) {
  static $fetched = array(); // We don't need to fetch every time.

  // Use batch directory.
  if (!is_dir(PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY)) {
    if (!mkdir(PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY)) {
      watchdog('project_dependency', 'Failed to create PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY %directory', array('%directory' => PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY));
      return FALSE;
    }
  }

  // Checkout release.
  $release['tag'] = escapeshellcmd($release['tag']);
  $release['uri'] = escapeshellcmd($release['uri']);
  $url = escapeshellarg(PROJECT_DEPENDENCY_UPSTREAM_REPOSITORY_PATH . '/' . $release['uri'] . '.git');
  $tag = escapeshellarg($release['tag']);
  $directory = $release['uri'];

  $work_tree = PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY . "/$directory";
  $git_dir = "$work_tree/.git";
  $esc_git_dir = escapeshellarg($git_dir);
  $work_tree = escapeshellarg($work_tree);
  // If the git directory doesn't exist, then clone into it.
  if (!is_dir($git_dir)) {
    $command = "git clone $url $work_tree 2>&1";
    exec($command, $output, $status);
    if ($status) {
      watchdog('project_dependency', 'Failed to execute git command %cmd; output=%output', array('%cmd' => $command, '%output' => implode(' ', $output)));
      return FALSE;
    }
  }
  // But if it did already exist
  else if (empty($fetched[$release['uri']])) {
    $command = "git --work-tree $work_tree --git-dir $esc_git_dir fetch 2>&1";
    exec($command, $output, $status);
    if ($status) {
      watchdog('project_dependency', 'Failed to execute git command %cmd; output=%output', array('%cmd' => $command, '%output' => implode(' ', $output)));
      return FALSE;
    }
  }
  $fetched[$release['uri']] = TRUE;

  // We need to make sure the directory is clean, so reset and clean
  // git reset --hard
  exec("git --work-tree $work_tree --git-dir $esc_git_dir reset --hard 2>&1", $output, $status);
  // git clean -fd
  exec("git --work-tree $work_tree --git-dir $esc_git_dir clean -fd 2>&1", $output, $status);
  // Make sure we don't have an existing branch by this name, or it won't be updated properly
  exec("git --work-tree=$work_tree --git-dir $esc_git_dir branch -D $tag 2>&1", $output, $status);
  // git checkout <identifier>
  $command = "git --work-tree=$work_tree --git-dir $esc_git_dir checkout $tag 2>&1";
  exec($command, $output, $status);
  if ($status) {
    watchdog('project_dependency', 'Failed to execute git command %cmd; output=%output', array('%cmd' => $command, '%output' => implode(' ', $output)));
    return FALSE;
  }

  // Scan checkout for .info files and create a list in the same format that
  // project release uses, so that standard API functions can be used.

  // Note that DRUPAL_PHP_FUNCTION_PATTERN croaks in file_scan_directory()
  // "WD php: Warning: ereg(): REG_ERANGE in file_scan_directory()"
  // $function_pattern = DRUPAL_PHP_FUNCTION_PATTERN;
  $function_pattern = '[a-zA-Z][a-zA-Z0-9_]*';
  $profile_info_files = file_scan_directory(PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY . '/' . $directory, '^' . $function_pattern . '\.(theme|profile)$');
  // If any profile or theme files are found, skip this release entirely.
  if (!empty($profile_info_files)) {
    return FALSE;
  }
  $files = file_scan_directory(PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY . '/' . $directory, '^' . $function_pattern . '\.info$');
  $info_files = array();
  foreach ($files as $file) {
    $info_files[] = $file->filename;
  }
  $info = project_dependency_info_parse($info_files);

  $dependencies = array();
  foreach ($info as $module => $module_info) {
    $dependencies[$module] = $module_info['dependencies'];
  }

  // Clear previous records for the release.
  project_dependency_info_package_clear($release['nid']);

  $components_in_release = array_keys($info);
  // Store the list of components contained by the project.
  $component_info = project_dependency_info_package_list_store($release['nid'], $info);
  foreach ($component_info as $component => $item) {
    $new_dependencies = array();
    if (is_array($info[$component]['dependencies'])) {
      foreach ($info[$component]['dependencies'] as $new_dependency) {
        $new_dependencies[$new_dependency] = array('external' => !in_array($new_dependency, $components_in_release));
      }
    }
    project_dependency_info_package_dependencies_store($item['release_nid'], $item['component_id'], $new_dependencies);
  }

  return $dependencies;
}


/**
 * Build dependency information for a single release node if it meets the
 * minimum Drupal major version API requirement.
 *
 * @param $shortname
 *   The shortname of the project.
 * @param $version
 *   The version of the project
 *
 * @return
 *   An array of dependency information or FALSE on failure.
 */
function project_dependency_process_release($shortname, $version) {
  // Load the project.
  $pid = project_get_nid_from_uri($shortname);
  $query = 'SELECT nid,tag FROM {project_release_nodes} WHERE pid = %d AND version = "%s"';
  $release = db_fetch_array(db_query($query, $pid, $version));
  $release['uri'] = $shortname;

  $node = node_load($release['nid']);
  if (empty($node)) {
    return FALSE;
  }
  $core_api_term = project_dependency_info_core_api($node);
  if (empty($core_api_term)) {
    return FALSE;
  }

  $core_api = project_dependency_core_api_from_term($core_api_term->name);
  if ($core_api < PROJECT_DEPENDENCY_MINIMUM_DRUPAL_VERSION) {
    return FALSE;
  }

  $dependencies = project_dependency_info_batch_process_release($release);
  return $dependencies;
}

/**
 * Find all dependencies (as release nids) of a given release nid and optional
 * components.
 *
 * If $depending_components are provided, only their dependencies will be
 * returned, rather than the dependencies of every component in the release.
 *
 * OK, I'm sorry this is a mindbending function. I've tried to compensate with
 * lots of comments, but suspect it's nowhere near adequate.
 *
 * This is a recursive, accumulative function, gathering the results we need in
 * $dependency_releases.
 *
 * @param $release_nid
 *   nid of the release node.
 * @param $depending_components
 *   an array of component names, like
 *   array('views_ui', 'token', 'page_manager')
 * @param $dependency_releases (by reference)
 *   The array of release nids that are depended upon by the given components.
 * @param $recursion_level
 *   Recursion level of this call to prevent infinite recursion on some
 *   database problem.
 *
 * Same as  project_dependency_info_process_dependencies($rid, $modules[$module], $dependencies);
 */
function project_dependency_get_external_component_dependencies($release_nid, $depending_components, &$dependency_releases, $recursion_level = 0) {
  // Limit this to 4 levels for now, as we probably don't know of anything
  // that needs recursion_level > 2.
  if ($recursion_level > 4) {
    return;
  }
  // Get the release node and project node of the *depending* components.
  $release_node = node_load($release_nid);
  $project_node = node_load($release_node->project_release['pid']);

  // We don't need to study the dependencies of the drupal project.
  if ($project_node->project['uri'] == 'drupal') {
    return;
  }

  // Get the api_term of the *depending* components.
  if (!($api_term = project_dependency_info_core_api($release_node))) {
    watchdog('project_dependency', 'ERROR: No core release API term found for release nid=%nid', array('%nid' => $release_nid));
    return;
  }

  // If no list of depending components was passed in, create a list of all
  // components in this release.
  if (empty($depending_components)) {
    $query = 'SELECT name FROM {project_dependency_component} WHERE release_nid = %d';
    $result = db_query($query, $release_nid);
    while ($row = db_fetch_array($result)) {
      $depending_components[] = $row['name'];
    }
  }
  // Now we have a list of component names for which we need to find
  // external dependencies. Walk through them and for each one, find the
  // external (not in this release) dependencies.
  foreach ($depending_components as $depending_component) {
    // Get a list of keyed arrays the describe external dependencies.
    $external_dependencies = project_dependency_find_external_dependencies($release_nid, $depending_component);

    if (empty($external_dependencies)) {
      continue;
    }

    // Walk through the external dependencies to find releases they are shipped
    // in.
    foreach ($external_dependencies as $component_info) {
      $dependency_component = $component_info['dependency'];

      // Find the releases that component was shipped in.
      // We could hope that it would only be in one package/rid,
      // but there are probably cases (refactoring) where that is not
      // the case.
      $possible_dependent_releases = project_dependency_info_releases_for_component_get($dependency_component, $api_term->tid);
      // We choose to use the project of the most recently created release
      // node. This is not guaranteed, but is a heuristic. Is there another?
      $first_dependent_release = reset($possible_dependent_releases);
      $dependent_pid = $first_dependent_release['pid'];
      $recommended_release = project_release_get_current_recommended($dependent_pid, $api_term->tid);

      // If we found a recommended release for this project and that release
      // also contains this component, accept it.
      if (!empty($recommended_release) && !empty($possible_dependent_releases[$recommended_release->nid])) {
        // Accept the recommended release for the project used.
        $release = $possible_dependent_releases[$recommended_release->nid];
      }
      else {
        // The recommended release does not contain the component, so just
        // punt and accept the first release that was found.
        $release = $first_dependent_release;
      }
      $release['component'] = $dependency_component;
      $dependency_releases[$dependency_component] = $release;
      // Recurse to find dependencies of the discovered dependency, but only
      // for the component which discovered the dependent release.
      project_dependency_get_external_component_dependencies($release['release_nid'], array($dependency_component), $dependency_releases, $recursion_level + 1);
    }
  }
}


/**
 * Given a release nid and component name, find all external dependencies for
 * that component.
 * @param $release_nid
 *   Release nid.
 * @param $component_name
 *   Component name.
 *
 * @return
 *   An array of component information.
 */
function project_dependency_find_external_dependencies($release_nid, $component_name) {
  $rows = array();
  $query = 'SELECT DISTINCT(pdc.component_id), name AS depender, dependency, release_nid FROM {project_dependency_dependency} pdd INNER JOIN {project_dependency_component} pdc
            ON pdd.component_id = pdc.component_id WHERE  pdc.name = "%s" AND pdc.release_nid = %d AND pdd.external = 1';
  $result = db_query($query, $component_name, $release_nid);
  while($row = db_fetch_array($result)) {
    $rows[] = $row;
  }
  return $rows;
}
