<?php
/**
 * @file
 * Provide Drush integration for release building and dependency building.
 */

/**
 * Implements hook_drush_command().
 */
function project_dependency_drush_command() {
  $items = array();

  $items['project-dependency-process-project'] = array(
    'description' => 'Build dependencies for a given project shortname.',
    'callback' => 'drush_project_dependency_process_project',
    'arguments' => array(
      'shortname' => 'Project shortname/uri',
    ),
    'aliases' => array('pdpp'),
  );
  $items['project-dependency-process-version'] = array(
    'description' => 'Build dependencies for a given project version (primarily for debugging).',
    'callback' => 'drush_project_dependency_process_version',
    'arguments' => array(
      'shortname' => 'Project shortname/uri',
      'version' => 'Version (like 7.x-3.x-dev)',
    ),
    'aliases' => array('pdpv'),
  );
  $items['project-dependency-process-all-projects'] = array(
    'description' => 'Build dependency records for all projects. This will take a long time',
    'callback' => 'drush_project_dependency_process_all_projects',
    'aliases' => array('pdap'),
  );
  $items['project-dependency-show-dependencies'] = array(
    'description' => 'Show the dependencies of the given release node',
    'callback' => 'drush_project_show_dependencies',
    'arguments' => array(
      'shortname' => 'Shortname of the project (uri)',
      'version' => 'Version of the project (like "7.x-1.x-dev"',
    ),
    'aliases' => array('pdsd'),
  );

  return $items;
}



/**
 * Update and save all dependencies for the project named.
 *
 * @param $shortname
 *   Any number of project shortnames to be processed
 */
function drush_project_dependency_process_project() {
  module_load_include('inc', 'project_dependency', 'project_dependency.drupal');

  $args = func_get_args();

  foreach ($args as $shortname) {
    drush_print("Processing dependencies for $shortname");

    // Get all versions for a project and build dependencies for them.
    $pid = project_get_nid_from_uri($shortname);

    // Find all releases for a project.
    $query = 'SELECT nid,tag,version FROM {project_release_nodes} WHERE pid = %d ORDER BY nid DESC';
    $result = db_query($query, $pid);
    while ($release_info = db_fetch_array($result)) {
      $dependencies = project_dependency_process_release($shortname, $release_info['version']);
      if ($dependencies === FALSE) {
        drush_print("Skipped (or git failed) $shortname: {$release_info['version']}");
        continue;
      }
      drush_print_r("Processed $shortname: {$release_info['version']}");
    }
  }
}


/**
 * Process a single release node (version)
 * @param $shortname
 *   The project shortname (uri)
 * @param $version
 *   The version string (like 7.x-3.x-dev)
 */
function drush_project_dependency_process_version($shortname, $version) {
  module_load_include('inc', 'project_dependency', 'project_dependency.drupal');
  drush_print("Processing dependencies for $shortname: $version");

  $dependencies = project_dependency_process_release($shortname, $version);
  if ($dependencies === FALSE) {
    drush_print("Skipped $shortname: $version");
    return;
  }
  drush_print_r($dependencies);
}

/**
 * Drush handler for processing all projects.
 *
 * Harvests and stores dependency information for all projects.
 * Just let this one run.
 */
function drush_project_dependency_process_all_projects() {
  ini_set('memory_limit', -1);
  module_load_include('inc', 'project_dependency', 'project_dependency.drupal');

  $total = $to_do = db_result(db_query('SELECT COUNT(*) FROM {project_release_nodes}'));
  $done = 0;

  $query = 'SELECT prn.nid, prn.version, p.uri AS shortname FROM {project_release_nodes} prn INNER JOIN {project_projects} p ON p.nid = prn.pid ORDER BY shortname, version';
  $result = db_query($query);
  $count = $skipped = 0;
  while ($release_info = db_fetch_array($result)) {
    $dependencies = project_dependency_process_release($release_info['shortname'], $release_info['version']);
    if ($dependencies === FALSE) {
      $skipped++;
    }
    $pct = round($done * 100/$total, 2);
    drush_print("$pct% done. Done: $done; Skipped: $skipped; Remaining: $to_do  ({$release_info['shortname']}:{$release_info['version']} nid={$release_info['nid']})");
    $done++;
    $to_do--;
  }
}

/**
 * Show dependencies for a given project + version
 *
 * @param $shortname
 *   Project shortname (uri) like 'views'
 * @param unknown_type $version
 *   Project version, like '7.x-1.1'
 */
function drush_project_show_dependencies($shortname, $version) {
  module_load_include('inc', 'project_dependency', 'project_dependency.drupal');

  // Find the release nid for the given shortname + version
  $pid = project_get_nid_from_uri($shortname);
  $nid = db_result(db_query('SELECT nid FROM {project_release_nodes} WHERE pid = %d AND version = "%s"', $pid, $version));
  if (empty($nid)) {
    drush_print_r("Failed to load a release for $shortname:$version");
    return;
  }
  $dependent_releases = project_dependency_get_external_release_dependencies($nid);

  drush_print_r($dependent_releases);
}
